#include <fstream>
#include <cstring>
#include <cstdio>
#include <cmath>
#include <csignal>
#include "rplot.h"
#include "mapplot.h"

extern "C" {
#include "chil.h"
}

#define DEBUG cerr << "Reached line " << __LINE__ << endl;
//#define DEBUG /* */

#define ERROR(msg,status) cerr << "Error: " << (msg) << "\n"; exit(status)
#define FLUSH(msg) cerr << (msg) << ": flushing figure\n"; flush_and_exit()
#define WARNING(msg) cerr << "Warning: " << (msg) << "\n"
#define MINIMUM(a,b) (((a) > (b)) ? (b) : (a))
#define MAXIMUM(a,b) (((a) < (b)) ? (b) : (a))
#define NO_SIGNAL -999.00
#define ANGLE_SCALE 0.000290888
#define PI_OVER_180 0.01745329
#define MAX_PLOTS 16

#ifndef DEFAULT_X_FONT_SERVER
#define DEFAULT_X_FONT_SERVER "unix/:7100"
#endif

enum FILE_FORMAT {
  FORMAT_POSTSCRIPT,
  FORMAT_PNG
};

/*** Global variables ***/
ChilboltonData data;
int nplots = 0;
RFigure *figure[MAX_PLOTS];
RAxes *axes[MAX_PLOTS], *cbar;
RPalette colormap;
RFontLoader *font_loader = NULL;
ostream *os = &cout;
ofstream ofs[MAX_PLOTS];
bool oneplot = false;

FILE_FORMAT format=FORMAT_PNG;

/*** Functions ***/
int
Usage()
{
  CHILUsage(&data);
  CHILPrintf("\n"
	     "Plotting options:\n"
	     "    -hourrange hr1 hr2  set hour axis range\n"
	     "     -duration hrs      set duration of cloud plots (overridden by `-hourrange')\n"
	     "  -fixduration mins     set duration of FIX plots\n"
	     "  -heightrange ht1 ht2  set height axis range\n"
	     "       -rrange x1 x2    set horizontal axis range for RHIs\n"
	     "       -Zrange Z1 Z2    set the Z colour scale range\n"
	     "       -vrange v1 v2    set the v colour scale range\n"
	     "   -PHIDPrange p1 p2    set the PHIDP colour scale range\n"
	     "     -LDRrange l1 l2    set the LDR colour scale range\n"
	     "          -mag f        set the magnification factor (default 1)\n"
	     "      -maxsize d        scale down image if panel is larger than d pixels\n"
	     "          -png          set image format to portable network graphics (default)\n"
	     "     -ps, -eps          set image format to encapsulated postscript\n"
	     "     -fontsize fs       set pixel size of font\n"
	     "      -oneplot          produce only one plot and send it to standard output\n"
	     "       -prefix dir      write the plots in specified directory\n"
	     "    -locations file     read locations to plot on PPIs from file\n"
	     "    -coastline file     read coastline to plot on PPIs from file\n"
	     "          -xfs server   specify the X font server (default "
	     DEFAULT_X_FONT_SERVER
	     ")\n");
  CHILOptions(&data);
  exit(0);
}

void
flush_and_exit()
{
  int i;
  if (font_loader) delete font_loader;

  for (i = 0; i < nplots; i++) {
    if (format == FORMAT_PNG && figure[i]) {
      ostream *os = &ofs[i];
      figure[i]->write_png(os);
    }
    ofs[i].close();

    /* Must delete these objects in the right order for some unknown reason -
       oh, the joy of C++ */
    if (axes[i]) delete axes[i];
    if (figure[i]) delete figure[i];
  }
  exit(0);
}

void
flush_on_signal(int signal)
{
  FLUSH("Interupt received");
}


/*** Main program ***/
int
main(int argc,char **argv)
{

  ChilboltonFile *chil_file = &(data.chil_file);
  RasterInfo *raster_info = &(data.raster_info);
  Ray *ray = &(data.ray);

  RFont *medium_font;

  double starthour=0, endhour=24, minheight=0, maxheight=12,dhour=1.0;
  double minr=0, maxr=95;
  float decimalhour, lastdecimalhour=-1, duration=6.0, fixduration=10.0;
  float correct_range[1024];
  bool cloud_data, setrrange=false;
  uchar sweep_mode;
  char datestr[128], titlestr[512], ofilestr[512];
  char namestr[128], tmpstr[128];
  char *prefix = ".", *scantype = "???";
  char *suffix = "png";
  char *mapfile = NULL;
  char *locfile = NULL;
  unsigned short int ifield[MAX_PLOTS];
  int daymod = 0;
  int font_size = 12;

  int status, iarg = 0;
  int cloud_width = 650, cloud_height = 230; /* default size for cloud plots */
  int width = 500,height = 500;
  int fig_width, fig_height;
  float x_scale = 1, y_scale = 1;
  float x_mag = 1, y_mag = 1;
  int max_size = 0;
  int i, iray = 0;
  int x_offset = 50, y_offset = 40;
  float minx = -90, maxx = 90, miny = -90, maxy = 90;
  float angle = 0, dangle = 0, bestangle = 0;
  float firsthour = -1;

  unsigned short hour, minute, second;

  bool bydate=false, sethourrange=false, setheightrange=false;
  bool quiet = false, plotdate = true;

  /* Default field ranges for scanning data */
  float interval[MAX_FIELDS], maxval[MAX_FIELDS], minval[MAX_FIELDS];
  /* Default field ranges for cloud data */
  float interval_c[MAX_FIELDS], maxval_c[MAX_FIELDS], minval_c[MAX_FIELDS];

  float intervalue[MAX_PLOTS], maxvalue[MAX_PLOTS], minvalue[MAX_PLOTS];

  char *xfs = DEFAULT_X_FONT_SERVER;

  status = CHILInit(argc, argv, &data);

  if (argc < 2) {
    Usage();
  }

  for (i = 0; i < MAX_FIELDS; i++) {
    minval[i] = -15;
    maxval[i] = 15;
    interval[i] = 5;

    minval_c[i] = -15;
    maxval_c[i] = 15;
    interval_c[i] = 5;
  }
  minval[ZH] = -30; maxval[ZH] = 50; interval[ZH] = 10;
  minval[ZDR] = -1; maxval[ZDR] = 4; interval[ZDR] = 1;
  minval[PHIDP] = -10; maxval[PHIDP] = 40; interval[PHIDP] = 10;
  minval[LDR] = -40; maxval[LDR] = 0; interval[LDR] = 10;
  minval[V1] = -7.5; maxval[V1] = 7.5; interval[V1] = 2.5;
  minval[WIDTH_H] = 0; maxval[WIDTH_H] = 5; interval[WIDTH_H] = 1;
  minval[WIDTH_V] = 0; maxval[WIDTH_V] = 5; interval[WIDTH_V] = 1;

  minval_c[ZH] = -50; maxval_c[ZH] = 10; interval_c[ZH] = 10;
  minval_c[V2] = -5; maxval_c[V2] = 5; interval_c[V2] = 1;
  minval_c[V_STD] = 0; maxval_c[V_STD] = 1; interval_c[V_STD] = 0.2;

  /* If looking at all rasters from many files, want 0-24 UTC plots
     by default */
  if (strcmp(argv[1], "-all") == 0) {
    sethourrange=true;
  }

  /* Look for plot-related arguments */
  while (++iarg < argc) {
    if (strcmp(argv[iarg],"-date") == 0) {
      bydate=true;
    }
    else if (strcmp(argv[iarg],"-hourrange") == 0) {
      if (argc == iarg+1 || sscanf(argv[++iarg],"%lf",&starthour) < 1) {
	ERROR("Failed to read start hour as float",1);
      }
      if (argc == iarg+1 || sscanf(argv[++iarg],"%lf",&endhour) < 1) {
	ERROR("Failed to read end hour as float",1);
      }
      sethourrange=true;
    }
    else if (strcmp(argv[iarg],"-heightrange") == 0) {
      if (argc == iarg+1 || sscanf(argv[++iarg],"%lf",&minheight) < 1) {
	ERROR("Failed to read minimum height as float",1);
      }
      if (argc == iarg+1 || sscanf(argv[++iarg],"%lf",&maxheight) < 1) {
	ERROR("Failed to read maximum height as float",1);
      }
      setheightrange=true;
    }
    else if (strcmp(argv[iarg],"-rrange") == 0) {
      if (argc == iarg+1 || sscanf(argv[++iarg],"%lf",&minr) < 1) {
	ERROR("Failed to read minimum r as float",1);
      }
      if (argc == iarg+1 || sscanf(argv[++iarg],"%lf",&maxr) < 1) {
	ERROR("Failed to read maximum r as float",1);
      }
      setrrange=true;
    }
    else if (strcmp(argv[iarg],"-Zrange") == 0) {
      if (argc == iarg+1 || sscanf(argv[++iarg],"%f",&minval[ZH]) < 1) {
	ERROR("Failed to read minimum Z as float",1);
      }
      if (argc == iarg+1 || sscanf(argv[++iarg],"%f",&maxval[ZH]) < 1) {
	ERROR("Failed to read maximum Z as float",1);
      }
      if (minval[ZH] >= maxval[ZH]) {
	ERROR("Minimum Z must be less than maximum Z",1);
      }
      minval_c[ZH] = minval[ZH];
      maxval_c[ZH] = maxval[ZH];
    }
    else if (strcmp(argv[iarg],"-vrange") == 0) {
      if (argc == iarg+1 || sscanf(argv[++iarg],"%f",&minval[V1]) < 1) {
	ERROR("Failed to read minimum v as float",1);
      }
      if (argc == iarg+1 || sscanf(argv[++iarg],"%f",&maxval[V1]) < 1) {
	ERROR("Failed to read maximum v as float",1);
      }
      if (minval[V1] >= maxval[V1]) {
	ERROR("Minimum v must be less than maximum v",1);
      }
      minval[V2] = minval[V1];
      maxval[V2] = maxval[V1];
      minval_c[V1] = minval[V1];
      maxval_c[V1] = maxval[V1];
      minval_c[V2] = minval[V1];
      maxval_c[V2] = maxval[V1];
    }
    else if (strcmp(argv[iarg],"-widthrange") == 0) {
      if (argc == iarg+1 || sscanf(argv[++iarg],"%f",&minval[WIDTH]) < 1) {
	ERROR("Failed to read minimum width as float",1);
      }
      if (argc == iarg+1 || sscanf(argv[++iarg],"%f",&maxval[WIDTH]) < 1) {
	ERROR("Failed to read maximum width as float",1);
      }
      if (minval[WIDTH] >= maxval[WIDTH]) {
	ERROR("Minimum width must be less than maximum width",1);
      }
      minval[WIDTH_H] = minval[WIDTH_V] = minval[WIDTH];
      maxval[WIDTH_H] = maxval[WIDTH_V] = maxval[WIDTH];
      minval_c[WIDTH_H] = minval_c[WIDTH_V] = minval[WIDTH];
      maxval_c[WIDTH_H] = maxval_c[WIDTH_V] = maxval[WIDTH];
    }
    else if (strcmp(argv[iarg],"-vstdrange") == 0) {
      if (argc == iarg+1 || sscanf(argv[++iarg],"%f",&minval[V_STD]) < 1) {
	ERROR("Failed to read minimum vstd as float",1);
      }
      if (argc == iarg+1 || sscanf(argv[++iarg],"%f",&maxval[V_STD]) < 1) {
	ERROR("Failed to read maximum vstd as float",1);
      }
      if (minval[V_STD] >= maxval[V_STD]) {
	ERROR("Minimum vstd must be less than maximum vstd",1);
      }
      minval_c[V_STD] = minval[V_STD];
      maxval_c[V_STD] = maxval[V_STD];
    }
    else if (strcmp(argv[iarg],"-PHIDPrange") == 0) {
      if (argc == iarg+1 || sscanf(argv[++iarg],"%f",&minval[PHIDP]) < 1) {
	ERROR("Failed to read minimum PHIDP as float",1);
      }
      if (argc == iarg+1 || sscanf(argv[++iarg],"%f",&maxval[PHIDP]) < 1) {
	ERROR("Failed to read maximum PHIDP as float",1);
      }
      if (minval[PHIDP] >= maxval[PHIDP]) {
	ERROR("Minimum PHIDP must be less than maximum PHIDP",1);
      }
    }
    else if (strcmp(argv[iarg],"-LDRrange") == 0) {
      if (argc == iarg+1 || sscanf(argv[++iarg],"%f",&minval[LDR]) < 1) {
	ERROR("Failed to read minimum LDR as float",1);
      }
      if (argc == iarg+1 || sscanf(argv[++iarg],"%f",&maxval[LDR]) < 1) {
	ERROR("Failed to read maximum LDR as float",1);
      }
      if (minval[LDR] >= maxval[LDR]) {
	ERROR("Minimum LDR must be less than maximum LDR",1);
      }
    }
    else if (strcmp(argv[iarg],"-mag") == 0) {
      if (argc == iarg+1 || sscanf(argv[++iarg],"%f",&x_mag) < 1) {
	ERROR("Failed to read magnification value",1);
      }
      if (x_mag < 0.01) {
	x_mag = 0.01;
	CHILPrintf("Using magnification factor of %g\n", x_mag);
      }	
      else if (x_mag > 10) {
	x_mag = 10.0;
	CHILPrintf("Using magnification factor of %g\n", x_mag);
      }
      y_mag = x_mag;
    }
    else if (strcmp(argv[iarg],"-maxsize") == 0) {
      if (argc == iarg+1 || sscanf(argv[++iarg],"%d",&max_size) < 1) {
	ERROR("Failed to read maximum panel size",1);
      }
    }
    else if (strcmp(argv[iarg], "-fontsize") == 0) {
      if (argc <= iarg+1) {
	ERROR("Failed to read font size",1);
      }
      font_size = atoi(argv[++iarg]);
      if (font_size < 1) {
	ERROR("Font size must be a positive integer",1);
      }
    }
    else if ((strcmp(argv[iarg],"-ps") == 0) 
	     || (strcmp(argv[iarg],"-eps") == 0)) {
      format = FORMAT_POSTSCRIPT;
      suffix = "eps";
    }
    else if (strcmp(argv[iarg],"-png") == 0) {
      format = FORMAT_PNG;
    }
    else if (strcmp(argv[iarg],"-oneplot") == 0) {
      oneplot = true;
    }
    else if (strcmp(argv[iarg],"-duration") == 0) {
      if (argc > iarg+1) {
	duration=atof(argv[++iarg]);
      }
      else {
	ERROR("Duration not specified",1);
      }
    }
    else if (strcmp(argv[iarg],"-fixduration") == 0) {
      if (argc > iarg+1) {
	fixduration=atof(argv[++iarg]);
      }
      else {
	ERROR("Duration not specified",1);
      }
    }
    else if (strcmp(argv[iarg],"-prefix") == 0) {
      if (argc > iarg+1) {
	prefix = argv[++iarg];
      }
      else {
	ERROR("Prefix not specified",1);
      }
    }
    else if (strcmp(argv[iarg],"-coastline") == 0) {
      if (argc > iarg+1) {
	mapfile = argv[++iarg];
      }
      else {
	ERROR("Coastline file not specified",1);
      }
    }
    else if (strcmp(argv[iarg],"-locations") == 0) {
      if (argc > iarg+1) {
	locfile = argv[++iarg];
      }
      else {
	ERROR("Location file not specified",1);
      }
    }
    else if (strcmp(argv[iarg],"-xfs") == 0) {
      if (argc > iarg+1) {
	xfs = argv[++iarg];
      }
      else {
	ERROR("X font server not specified",1);
      }
    }

  }

  quiet = data.quiet;

  try {

    /* Initialise general plot stuff */
    font_loader = new RFontLoader(xfs, format == FORMAT_PNG);
    font_loader->verbose(data.verbose);
    medium_font = font_loader->load(font_size,
				    string("helvetica"),
				    string("medium"),
				    string("r"),
				    string("adobe"));

    create_jet_palette(&colormap,64);

    if (bydate) {
      sethourrange = true;
    }
    MAPSetOrigin(CHILLatitude(raster_info), CHILLongitude(raster_info));

    if (mapfile) {
      if (MAPLoadCoastline(mapfile)) {
	CHILPrintf("Could not load coastline file %s\n", mapfile);
	mapfile = NULL;
      }
    }

    if (locfile) {
      if (MAPLoadLocations(locfile)) {
	CHILPrintf("Could not load locations file %s\n", locfile);
	locfile = NULL;
      }
    }

    signal(SIGINT,flush_on_signal);
    signal(SIGTERM,flush_on_signal);

    if (status == CHIL_OK)
      {
	while (1)
	  {
	    int failures = 0, maxfailures = 10;

	    while ((status = CHILFetchRay(&data)) != CHIL_OK)
	      {
		if (status != CHIL_ENDOFBLOCK) {
		  break;
		}
		if ((++failures) == maxfailures)
		  {
		    CHILPrintf("Aborting: %d failed calls to CHILFetchRay - empty scans?\n");
		    break;
		  }
	      }
	    if (status != CHIL_OK)
	      {
		break;
	      }

	    /*** ALL DATA ***/
	    cloud_data = raster_info->cloud_data;
	    iray = 0;

	    for (i = 0; i < ray->ntotal; i++) {
	      ifield[i] = i;
	    }
	    nplots = ray->ntotal;
	    if (oneplot) {
	      nplots = 1;
	    }
	    if (nplots > data.maxfields) {
	      nplots = data.maxfields;
	    }

	    firsthour = ray->utc;
	    if (cloud_data || firsthour != 0) {
	      hour = (unsigned short) firsthour;
	      minute = (unsigned short) (60.0 * (firsthour - (float) hour));
	      second = (unsigned short) (3600.0 * ((firsthour - (float) hour)
						   - minute/60.0));
	    }
	    else {
	      hour = raster_info->hour;
	      minute = raster_info->minute;
	      second = raster_info->second;
	    }

	    x_scale = 1; y_scale = 1;

	    /* Must remember sweep mode */
	    sweep_mode = raster_info->sweep_mode;
	    for (i = 0; i < ray->ngates; i++) {
	      correct_range[i] = ray->range[i] + ray->range_offset*0.001;
	    }
	    daymod = 0;
	    lastdecimalhour = -1;
	    if (cloud_data) {
	      /*** CLOUD DATA ***/
	      width = cloud_width;
	      height = cloud_height;
	      sprintf(namestr, "%s %d GHz radar (%s)",
		      CHILLocation(raster_info),
		      (int) raster_info->frequency, 
		      CHILSystemName(raster_info, NULL));
	      sprintf(titlestr, "<B>%s", namestr);
	      if (!sethourrange) {
		starthour = raster_info->hour 
		  + (((float) raster_info->minute)/60.0) 
		  + (((float) raster_info->second)/3600.0);
		endhour = starthour+duration;
	      }
		
	      if (endhour <= starthour) {
		FLUSH("Error: End hour is less than start hour!");
	      }
	      for (i = 0; i < nplots; i++) {
		minvalue[i] = minval_c[ray->field_type[ifield[i]]];
		maxvalue[i] = maxval_c[ray->field_type[ifield[i]]];
		intervalue[i] = interval_c[ray->field_type[ifield[i]]];
	      }
	      minx = starthour;
	      maxx = endhour;
	      miny = minheight;
	      maxy = maxheight;
	    }
	    else {
	      /*** SCANNING DATA ***/
	      angle = -1000;
	      sprintf(namestr,"%s %d GHz radar",
		      CHILLocation(raster_info),
		      (int) raster_info->frequency);
	      CHILGetDate(datestr,ray);

	      switch (sweep_mode) {
	      case UF_RHI: case UF_COPLANE:
		scantype = "RHI";
		y_scale = 4;
		if (raster_info->system_id == SINGAPORE) {
		  x_scale = 4;
		}
		miny = minheight;
		maxy = maxheight;
		if (setrrange) {
		  minx = minr;
		  maxx = maxr;
		}
		else {
		  minx = correct_range[0]*cos(raster_info->max_angle
					      * ANGLE_SCALE);
		  maxx = MAXIMUM(correct_range[ray->ngates-1],90.0/x_scale);
		  if (maxx < minx + 90.0/x_scale) {
		    if (maxx < 90.0/x_scale) {
		      minx = 0;
		      maxx = 90.0/x_scale;
		    }
		    else {
		      minx = maxx - 90.0/x_scale;
		    }
		  }
		}
		sprintf(titlestr,"<B>%s</B> %s "
			"%02hu:%02hu:%02hu %s %04hu/%03hd/%02hu Azim %5.1f<SUP>o",
			namestr, datestr,
			hour, minute, second,
			scantype,
			raster_info->file, raster_info->raster,
			chil_file->iscan+1,
			/* raster_info->init_angle/60.0-90.0 */
			ray->azimuth - 90.0);
		break;
	      case UF_PPI:
		scantype = "PPI";
		minx = miny = 100;
		maxx = maxy = -100;
		for (i = raster_info->min_angle;
		     i < raster_info->max_angle+60; i += 60) {
		  minx = MINIMUM(minx, correct_range[ray->ngates-1]	
				 * (-cos(((float) i) * ANGLE_SCALE)));
		  minx = MINIMUM(minx, correct_range[0]	
				 * (-cos(((float) i) * ANGLE_SCALE)));
		  maxx = MAXIMUM(maxx, correct_range[ray->ngates-1] 
				 * (-cos(((float) i) * ANGLE_SCALE)));
		  maxx = MAXIMUM(maxx, correct_range[0] 
				 * (-cos(((float) i) * ANGLE_SCALE)));
		  miny = MINIMUM(miny, correct_range[ray->ngates-1] 
				 * (sin(((float) i) * ANGLE_SCALE)));
		  miny = MINIMUM(miny, correct_range[0] 
				 * (sin(((float) i) * ANGLE_SCALE)));
		  maxy = MAXIMUM(maxy, correct_range[ray->ngates-1] 
				 * (sin(((float) i) * ANGLE_SCALE)));
		  maxy = MAXIMUM(maxy, correct_range[0] 
				 * (sin(((float) i) * ANGLE_SCALE)));
		}
		if (minx >= maxx-90) {
		  minx -= (90-maxx+minx)/2.0;
		  maxx += (90-maxx+minx)/2.0;
		}
		if (miny >= maxy) {
		  miny = -90.0; maxy = 90.0;
		}
		sprintf(titlestr,"<B>%s</B> %s "
			"%02hu:%02hu:%02hu %s %04hu/%03hd/%02hu Elev %4.1f<SUP>o",
			namestr, datestr,
			hour, minute, second,
			scantype,
			raster_info->file, raster_info->raster,
			(int) (chil_file->iscan+1),
			/* raster_info->init_angle/60.0 */
			ray->elevation);
		break;
	      default:
		scantype = "FIX";
		y_scale = 4;
		minx = 0;
		maxx = fixduration;
		miny = minheight;
		maxy = maxheight;
		x_scale = 100.0/fixduration;
		sprintf(titlestr,"<B>%s</B> %s "
			"%02hu:%02hu:%02hu %s %hu/%hd/%hu",
			namestr, datestr,
			hour, minute, second,
			scantype,
			raster_info->file, raster_info->raster,
			(int) (chil_file->iscan+1));
	      }
	      width = (int) ((maxx-minx) * x_scale * x_mag * 5);
	      height = (int) ((maxy-miny) * y_scale * y_mag * 5);
	      for (i=0;i<nplots;i++) {
		minvalue[i] = minval[ray->field_type[ifield[i]]];
		maxvalue[i] = maxval[ray->field_type[ifield[i]]];
		intervalue[i] = interval[ray->field_type[ifield[i]]];
	      }
	      minx = floor(minx);
	      miny = floor(miny);
	      maxx = ceil(maxx);
	      maxy = ceil(maxy);
	    }
	    /*** ALL DATA ***/
	    /* Check width & height do not exceed max_size */
	    if (max_size > 0) {
	      if (width > height) {
		if (width > max_size) {
		  int tmp_width = width;
		  width = max_size;
		  height = (max_size * height) / tmp_width;
		}
	      }
	      else {
		if (height > max_size) {
		  int tmp_height = height;
		  height = max_size;
		  width = (max_size * width) / tmp_height;
		}
	      }
	    }

	    fig_width = width + x_offset + 80;
	    fig_height = height + y_offset + 30;

	    if (!oneplot) {
	      if (!quiet) {
		cerr << "Writing";
	      }
	      for (i=0;i<nplots;i++) {
		if (cloud_data) {
		  sprintf(ofilestr,
			  "%s/%04hu%02hu%02hu_%u_%s.%s",
			  prefix, CHILTrueYear(ray->year),
			  ray->month, ray->day,
			  (unsigned int) raster_info->frequency,
			  CHILFieldName(ray, ifield[i], NULL),
			  suffix);
		}
		else {
		  sprintf(ofilestr,
			  "%s/%04hur%03hus%02hu_%02hu%02hu%02hu_%s_%u_%s.%s",
			  prefix,
			  raster_info->file, raster_info->raster, 
			  chil_file->iscan+1,
			  hour, minute, second,
			  scantype, (unsigned int) raster_info->frequency,
			  CHILFieldName(ray, ifield[i], NULL),
			  suffix);
		}
		ofs[i].open(ofilestr);
		os = &ofs[0];
		cerr << " " << ofilestr;
	      }
	      if (!quiet) {
		cerr << "\n";
	      }
	    }

	    if (format == FORMAT_POSTSCRIPT) {
	      for (i=0;i<nplots;i++) {
		if (oneplot) {
		  figure[i] = new RFigure(fig_width,fig_height,POSTSCRIPT,
					  os, namestr, 0.5);
		}
		else {
		  figure[i] = new RFigure(fig_width,fig_height,POSTSCRIPT,
					  &ofs[i], namestr, 0.5);
		}
	      }
	    }
	    else {
	      for (i=0;i<nplots;i++) {
		figure[i] = new RFigure(fig_width,fig_height,BITMAP_8BIT);
	      }
	    }
	    for (i=0;i<nplots;i++) {
	      figure[i]->clear();
	      figure[i]->axis(0,0,fig_width,fig_height);
	      figure[i]->font = medium_font;
	      cbar = new RAxes(figure[i], fig_width-26,
				  y_offset+0.5*height-80,
				  fig_width-10, y_offset+0.5*height+80);

	      cbar->colorbar(&colormap,VERTICAL);
	      cbar->box();
	      cbar->axis(0,minvalue[i],1,maxvalue[i]);
	      cbar->yticks(minvalue[i],intervalue[i],"%g",0);
	      sprintf(tmpstr,"%s (%s)", CHILFieldDescription(ray, ifield[i],
							     NULL),
		      CHILFieldUnits(ray, ifield[i], NULL));
	      cbar->ylabel(tmpstr);
	      delete cbar;

	      axes[i] = new RAxes(figure[i], x_offset, y_offset, 
				  x_offset + width, y_offset + height);
	      axes[i]->axis(minx, miny, maxx, maxy);
	      axes[i]->clipbox_to_axis();
	      if (cloud_data) {
		figure[i]->text(0.01, 0.0, titlestr, ANCHOR_BOTTOMLEFT);
	      }
	      else {
		axes[i]->text(maxx, maxy, titlestr, ANCHOR_BOTTOMRIGHT);
	      }
	    }
  
	    do
	      {
		decimalhour = ray->utc + 24*daymod;
		plotdate = (lastdecimalhour < 0 && ray->utc < 23.0);
		if (lastdecimalhour > decimalhour) {
		  ++daymod;
		  decimalhour += 24;
		  plotdate = true;
		}
		if (decimalhour > endhour) {
		  break;
		}
		else if (decimalhour < starthour) {
		  continue;
		}
		if (lastdecimalhour < 0) {
		  lastdecimalhour = decimalhour - 1.0/3600.0;
		}
		if (cloud_data)
		  {
		    /*** CLOUD DATA ***/
		    CHILGetDate(datestr,ray);
		    for (i = 0; i < nplots; i++) {
		      if (plotdate) {
			axes[i]->text(decimalhour,maxheight,datestr,
				      ANCHOR_BOTTOMLEFT);
		      }
		      axes[i]->vray(lastdecimalhour + ray->averays/3600.0,
				    decimalhour + ray->averays/1800.0,
				    correct_range, ray->data[i], ray->ngates,
				    &colormap, minvalue[i], maxvalue[i],
				    NO_SIGNAL);
		    }
		    if (plotdate) plotdate = false;
		  }
		else
		  {
		    /*** SCANNING DATA ***/
		    int j, k;
		    float x[4], y[4];
		    float tmpangle = 0;

		    switch (sweep_mode) {
		    case UF_RHI: case UF_COPLANE:
		      tmpangle = ray->elevation * PI_OVER_180;
		      if (angle < -999) {
			dangle = 0.2 * PI_OVER_180; // default:+/-0.2 degs
			bestangle = tmpangle;
		      }
		      else {
			dangle = 0.7 * (tmpangle - angle);
			bestangle = (tmpangle + angle) * 0.5;
		      }
		      angle = tmpangle;

		      break;
		    case UF_PPI:
		      if (angle < -999.0) {
			dangle = 0.2 * PI_OVER_180; // default:+/-0.2 degs
		      }
		      else {
			dangle = 0.7 * ((ray->azimuth) * PI_OVER_180
				       - angle);
		      }
		      angle = ray->azimuth * PI_OVER_180;
		      break;
		    }

		    for (i=0;i<ray->ngates-1;i++) {
		      if (*(ray->data[ifield[0]]+i) == NO_SIGNAL) {
			continue;
		      }
		      switch (sweep_mode) {
		      case UF_RHI: case UF_COPLANE:
			x[0] = correct_range[i]*cos(bestangle + dangle);
			y[0] = correct_range[i]*sin(bestangle + dangle);
			x[1] = correct_range[i+1]*cos(bestangle + dangle);
			y[1] = correct_range[i+1]*sin(bestangle + dangle);
			x[2] = correct_range[i+1]*cos(bestangle - dangle);
			y[2] = correct_range[i+1]*sin(bestangle - dangle);
			x[3] = correct_range[i]*cos(bestangle - dangle);
			y[3] = correct_range[i]*sin(bestangle - dangle);
			break;
		      case UF_PPI:
			x[0] = -correct_range[i]*cos(angle + dangle);
			y[0] = correct_range[i]*sin(angle + dangle);
			x[1] = -correct_range[i+1]*cos(angle + dangle);
			y[1] = correct_range[i+1]*sin(angle + dangle);
			x[2] = -correct_range[i+1]*cos(angle - dangle);
			y[2] = correct_range[i+1]*sin(angle - dangle);
			x[3] = -correct_range[i]*cos(angle - dangle);
			y[3] = correct_range[i]*sin(angle - dangle);
			break;
		      default:
			if (raster_info->time_stamping) {
			  x[0] = x[1] = (lastdecimalhour-firsthour)*60.0;
			  x[2] = x[3] = (decimalhour-firsthour+0.0003) * 60.0;
			}
			else {
			  x[0] = x[1] = iray * 0.25;
			  x[2] = x[3] = (iray+1) * 0.25;
			}
			y[0] = y[3] = correct_range[i];
			y[1] = y[2] = correct_range[i+1];
		      }
		      for (k = 0; k<nplots; k++) {
			j = (int) (((float) colormap.size())
				   * (*(ray->data[ifield[k]]+i)-minvalue[k])
				   / (maxvalue[k]-minvalue[k]));
			if (j < 0) j = 0;
			if (j > colormap.size()-1) j = colormap.size()-1;
			axes[k]->polygon(4, x, y, colormap.get(j));
		      }
		    }
		    iray++;
		  }
		lastdecimalhour=decimalhour;
	      }
	    while ((status = CHILFetchRay(&data)) == CHIL_OK);

	    CHILGetDate(datestr,ray);
	    for (i = 0; i < nplots; i++) {
	      axes[i]->box();
	      if (cloud_data) {
		if (endhour > 24.0) {
		  axes[i]->xtick(24,NULL,0,0,1);
		}
		dhour = (endhour-starthour > 6) ? floor((endhour-starthour)
							/ 6.0) :
		  (1.0+floor((endhour-starthour)/4.0))/2.0;
		axes[i]->xticks_time(0,dhour);
		axes[i]->yticks(0,1);
		axes[i]->xlabel("Time (UTC)");
		axes[i]->ylabel("Height (km)");
	      }
	      else {
		switch (sweep_mode) {
		case UF_RHI: case UF_COPLANE:
		  axes[i]->xticks(floor(minx*0.1)*10,10);
		  axes[i]->yticks(floor(miny*0.5)*2,2);
		  axes[i]->xlabel("Distance (km)");
		  axes[i]->ylabel("Height (km)");
		  break;
		case UF_PPI:
		  axes[i]->xticks(floor(minx*0.1)*10,10);
		  axes[i]->yticks(floor(miny*0.1)*10,10);
		  axes[i]->xlabel("Distance east (km)");
		  axes[i]->ylabel("Distance north (km)");
		  if (mapfile) {
		    figure[i]->clipbox(axes[i]->position());
		    MAPPlotCoastline(axes[i]);
		    figure[i]->clipbox(false);
		  }
		  if (locfile) {
		    figure[i]->clipbox(axes[i]->position());
		    MAPPlotLocations(axes[i]);
		    figure[i]->clipbox(false);
		  }
		  break;
		default:
		  axes[i]->xticks(floor(minx), 1);
		  axes[i]->yticks(floor(miny*0.5)*2,2);
		  axes[i]->xlabel("Approximate time (minutes)");
		  axes[i]->ylabel("Range (km)");
		}
	      }
	      if (format == FORMAT_PNG) {
		if (!oneplot) {
		  os = &ofs[i];
		}
		figure[i]->write_png(os);
	      }
	      delete axes[i]; axes[i] = NULL;
	      delete figure[i]; figure[i] = NULL;
	      ofs[i].close();

	    }
	    if (status == CHIL_ENDOFDATA)
	      {
		break;
	      }
	  }
	CHILEnd(&data);
     	delete font_loader; font_loader = NULL;
      }
  }
  catch  (const char *error_message) {
    cerr << error_message << endl;
    exit(1);
  }
  catch  (...) {
    cerr << "An error occurred" << endl;
    exit(1);
  }
}
