/* fftplot.cpp - plot some Doppler spectra */

#include <iostream>
#include <rplot.h>
#include <cstdio>
#include <cstring>

#include "cfft.h"

enum FILE_FORMAT {
  FORMAT_POSTSCRIPT,
  FORMAT_PNG
};

#ifndef DEFAULT_X_FONT_SERVER
#define DEFAULT_X_FONT_SERVER "unix/:7100"
#endif

int
get_date(char *datestr, int day, int month, int year)
{
  static char *months[] = {"Jan","Feb","Mar","Apr","May","June",
			   "July","Aug","Sept","Oct","Nov","Dec"};
  if (month < 1 || month > 12) {
    return 1;
  }
  else {
    if (year < 80) {
      year += 2000;
    }
    else if (year < 99) {
      year += 1900;
    }
    sprintf(datestr, "%d %s %d", day, months[month-1], year);
  }
  return 0;
}

void
Usage(char **argv)
{
  cerr << "Usage: " << argv[0] << " [options] profiles.fft > profile.png\n"
    "Options:\n"
    "  -heightrange ht1 ht2  set height axis range\n"
    "     -SNRrange r1 r2    set colour scale range for SNR (dB)\n"
    "     -reversev          negate the velocities\n"
    "       -system system   specify the name of the instrument\n"
    "      -profile n        plot profile n, -1 means plot most recent profile\n"
    "         -size <x>x<y>  set size of axes in pixels\n"
    "      -barsize <x>x<y>  set size of colourbar in pixels\n"
    "     -fontsize n        set pixel size of font\n"
    "       -system system   specify the name of the instrument\n"
    "          -png          set image format to portable network graphics (default)\n"
    "     -ps, -eps          set image format to encapsulated postscript\n"
    "          -xfs server   specify the X font server (default " 
    DEFAULT_X_FONT_SERVER ")\n";
  exit(0);
}

int
main(int argc, char **argv)
{
  CFFTData data;
  char *filename;
  int status;

  /* Plot properties */
  RFigure *figure;
  RAxes *axes, *cbar;
  RFontLoader *font_loader;
  RPalette colormap;
  RColor red(255, 0, 0), blue(0, 0, 255);

  FILE_FORMAT format=FORMAT_PNG;
  char *xfs = DEFAULT_X_FONT_SERVER;
  char *system = "Chilbolton 94 GHz Cloud Radar (Galileo)";
  int width = 256;
  int height = 230;
  int x_offset = 50, y_offset = 65;
  int fig_width, fig_height;
  int font_size = 12;
  float minheight = 0, maxheight = 12;
  float vfold = 5;
  int bar_width = 16, bar_height = 160;
  float minvalue = 5, maxvalue = 40, interval = 5;
  int iarg = 1;
  float range_offset = 0.0;
  float linescale = 100.0;
  int profile = 0;
  unsigned long flags = CFFT_USE_LOG_SNR
    | CFFT_REMOVE_NOISE | CFFT_REMOVE_SPIKE;
  bool reversev = false, no_bar = false;
  bool verbose = false, use_lines = false;

  while (iarg < argc && *argv[iarg] == '-') {
    if (strcmp(argv[iarg], "-png") == 0) {
      format = FORMAT_PNG;
    }
    else if ((strcmp(argv[iarg], "-ps") == 0)
	     || (strcmp(argv[iarg], "-eps") == 0)) {
      format = FORMAT_POSTSCRIPT;
    }
    if (strcmp(argv[iarg], "-lines") == 0) {
      use_lines = true;
      flags &= ~CFFT_REMOVE_NOISE;
      no_bar = true;      
    }
    else if (strcmp(argv[iarg], "-reversev") == 0) {
      reversev = true;
      flags |= CFFT_REVERSE_V;
    }
    else if (strcmp(argv[iarg], "-noclean") == 0) {
      flags &= ~CFFT_REMOVE_SPIKE;
    }
    else if (strcmp(argv[iarg],"-heightrange") == 0) {
      if (argc <= iarg+1 || sscanf(argv[++iarg],"%f",&minheight) < 1) {
	cerr << "Failed to read minimum height as float\n";
	exit(-1);
      }
      if (argc <= iarg+1 || sscanf(argv[++iarg],"%f",&maxheight) < 1) {
	cerr << "Failed to read maximum height as float";
	exit(-1);
      }
    }
    else if (strcmp(argv[iarg], "-system") == 0) {
      if (argc <= iarg+1) {
	cerr << "Failed to read system name\n";
	exit(-1);
      }
      system = argv[++iarg];
    }
    else if (strcmp(argv[iarg], "-SNRrange") == 0) {
      if (argc <= iarg+1 || sscanf(argv[++iarg],"%f",&minvalue) < 1) {
	cerr << "Failed to read minimum SNR\n";
	exit(-1);
      }
      if (argc <= iarg+1 || sscanf(argv[++iarg],"%f",&maxvalue) < 1) {
	cerr << "Failed to read maximum SNR\n";
	exit(-1);
      }
    }
    else if (strcmp(argv[iarg], "-size") == 0) {
      if (argc <= iarg+1 || sscanf(argv[++iarg],"%dx%d",
				   &width,&height) < 2) {
	cerr << "Failed to read size: should be in the form "
	  "<width>x<height>\n";
	exit(-1);
      }
    }
    else if (strcmp(argv[iarg], "-barsize") == 0) {
      if (argc <= iarg+1 || sscanf(argv[++iarg],"%dx%d",
				   &bar_width,&bar_height) < 2) {
	cerr << "Failed to read bar size: should be in the form "
	  "<width>x<height>\n";
	exit(-1);
      }
      if (bar_width < 1 || bar_height < 1) {
	no_bar = true;
      }
    }
    else if (strcmp(argv[iarg], "-fontsize") == 0) {
      if (argc <= iarg+1) {
	cerr << "Failed to read font size\n";
	exit(-1);
      }
      font_size = atoi(argv[++iarg]);
      if (font_size < 1) {
	cerr << "Font size must be a positive integer\n";
	exit(-1);
      }
    }
    else if (strcmp(argv[iarg], "-profile") == 0) {
      if (argc <= iarg+1) {
	cerr << "Failed to read number of required profile\n";
	exit(-1);
      }
      profile = atoi(argv[++iarg]);
    }
    else if (strcmp(argv[iarg], "-rangeoffset") == 0) {
      if (argc <= iarg+1) {
	cerr << "Failed to read range offset\n";
	exit(-1);
      }
      range_offset = atof(argv[++iarg]);
    }
    else if (strcmp(argv[iarg], "-linescale") == 0) {
      if (argc <= iarg+1) {
	cerr << "Failed to read line scale\n";
	exit(-1);
      }
      linescale = atof(argv[++iarg]);
    }
    else if (strcmp(argv[iarg],"-xfs") == 0) {
      if (argc > iarg+1) {
	xfs = argv[++iarg];
      }
      else {
	cerr << "X font server not specified";
	exit(-1);
      }
    }
    else {
      cerr << argv[iarg] << " not understood\n";
      Usage(argv);
    }
    ++iarg;
  }
  if (iarg > argc-1) {
    Usage(argv);
  }
	     
  if (no_bar) {
    fig_width = width + x_offset + 20;
  }
  else {
    fig_width = width + x_offset + bar_width + 64;
  }
  fig_height = height + y_offset + 30;

  filename = argv[iarg];

  cfft_verbose();
  cfft_set_processing_flags(&data, flags);
  cfft_set_range_offset(&data, range_offset);

  if (cfft_open(filename, &data) != CFFT_OK) {
    cerr << "Error reading " << filename << "\n";
    exit(2);
  }

  data.spikebleed = 1;
  cfft_set_noise_processing(&data, 50, 3);

  //  status = cfft_next_profile(&data);
  status = cfft_read_profile(&data, profile);

  cfft_process_spectra(&data);

  if (status) {
    cerr << "Could not read profile " << profile << "\n";
    exit(3);
  }
  else {
    try {
      /* Contact font server */
      font_loader = new RFontLoader(xfs, format == FORMAT_PNG);
      font_loader->verbose(verbose);

      /* One plot to standard output */
      /* Initialise plot */
      if (format == FORMAT_POSTSCRIPT) {
	figure = new RFigure(fig_width, fig_height, POSTSCRIPT,
			     &cout, "Chilbolton cloud profile", 0.5);
      }
      else {
	figure = new RFigure(fig_width, fig_height, BITMAP_8BIT);
      }
    
      figure->clear();
      figure->axis(0,0,fig_width, fig_height);
      figure->font = font_loader->load(font_size,
				       string("helvetica"),
				       string("medium"),
				       string("r"),
				       string("adobe"));
    
      axes = new RAxes(figure, x_offset, y_offset, 
		       x_offset + width, y_offset + height);
      axes->axis(-vfold, minheight, +vfold, maxheight);
      axes->clipbox_to_axis();

      create_jet_palette(&colormap,64);

      /* Plot the data */
      if (use_lines) {
	for (int i = data.ngates - 2; i >= 0; i -= 2) {
	  if (data.range[i] > minheight && data.range[i] < maxheight) {
	    axes->axis(-vfold, (minheight-data.range[i])*linescale,
		       vfold, (maxheight-data.range[i])*linescale);
	    if (i%10) {
	      axes->plot((float *)data.velocity,
			 (float *)data.spectra+i*data.nbins,
			 (int) data.nbins, blue);
	    }
	    else {
	      axes->plot((float *)data.velocity,
			 (float *)data.spectra+i*data.nbins,
			 (int) data.nbins, red);
	    }
	  }
	}
	axes->axis(-vfold, minheight, +vfold, maxheight);
      }
      else {
	for (int i = 0; i < data.ngates - 1; ++i) {
	  axes->hray(data.velocity, data.range[i], data.range[i+1],
		     data.spectra+i*data.nbins, data.nbins,
		     &colormap, minvalue, maxvalue, CFFT_NO_SIGNAL);
	}
      }

      /* Annotate */
      if (data.starttime >= 0.0) {
	char titlestr[512], tmpstr[512];
	get_date(tmpstr, data.day, data.month, data.year);
	sprintf(titlestr, "%s %02hu:%02hu UTC", tmpstr,
		data.hour, data.minute);
	axes->text(-vfold, maxheight, titlestr, ANCHOR_BOTTOMLEFT);
      }
      if (system) {
	figure->font = font_loader->load(font_size,
					    string("helvetica"),
					    string("bold"),
					    string("r"),
					    string("adobe"));
	figure->text(0.01, 0.0, system, ANCHOR_BOTTOMLEFT);
	figure->font = font_loader->load(font_size,
					    string("helvetica"),
					    string("medium"),
					    string("r"),
					    string("adobe"));
      }
      /* Finish up */
      axes->clipbox(false);
      axes->xticks(-vfold, 1,
		   NULL, DEFAULT_TICK_LENGTH, 0, 1);
      if (reversev) {
	axes->xlabel("Vertical velocity (m s<SUP>-1</SUP>, positive downwards)");
      }
      else {
	axes->xlabel("Vertical velocity (m s<SUP>-1</SUP>, positive upwards)");
      }
      axes->yticks(minheight, (float) ((int) (1+(maxheight-minheight)/14)),
		   NULL, DEFAULT_TICK_LENGTH, 0, 1);
      axes->ylabel("Height (km)");
      axes->box();

      if (!no_bar) {
	/* Add a colourbar */
	cbar = new RAxes(figure, (fig_width-10)-bar_width,
			 y_offset+0.5*(height-bar_height),
			 fig_width-10, y_offset+0.5*(height+bar_height));
	cbar->colorbar(&colormap,VERTICAL);
	cbar->box();
	cbar->axis(0, minvalue, 1, maxvalue);
	cbar->yticks(minvalue, interval, "%g",0);
	cbar->ylabel("Signal-to-noise ratio (dB)");
	delete cbar;
      }
      
      if (format == FORMAT_PNG) {
	figure->write_png(&cout);
      }
      delete axes;
      delete figure;
      delete font_loader;
    }
    catch  (const char * error_message) {
      cerr << error_message << endl;
      exit(1);
    }
  }
  
  cfft_close(&data);
  exit(0);
}
