#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <netcdf.h>

#include "ceil.h"
#include "nctools.h"
#include "files2stdin.h"
#include "version.h"

#define ERROR(msg,status) fprintf(stderr, "Error: %s\n", msg); exit(status)
#define WARNING(msg) fprintf(stderr, "Warning: %s\n", msg)

#define NO_SIGNAL -999.00

int ncstatus;
#define handle_error(a) if ((ncstatus = (a)) != NC_NOERR) { \
    fprintf(stderr, "NetCDF error on line %d: %s\n", __LINE__, \
            nc_strerror(ncstatus)); \
    exit(-1); \
}

void
add_attributes(int ncid, int fieldid, char *units,
	       char *units_html, char *long_name)
{
  if (units) {
    handle_error(nc_put_att_text(ncid, fieldid, "units",
				 strlen(units), units));
  }
  if (units_html) {
    handle_error(nc_put_att_text(ncid, fieldid, "units_html",
				 strlen(units_html), units_html));
  }
  if (long_name) {
    handle_error(nc_put_att_text(ncid, fieldid, "long_name",
				 strlen(long_name), long_name));
  }
}

void
add_scale_factor(int ncid, int fieldid, float betacal)
{
  float cal = betacal * 1.0e-7;

  handle_error(nc_put_att_float(ncid, fieldid, "scale_factor",
				NC_FLOAT, 1, &cal));
  nct_append_attribute(ncid, fieldid, "comments",
		       "To calibrate, multiply by the value "
		       "in the scale_factor attribute.", " ");
}

void
Usage(char **argv)
{
  fprintf(stderr, "Usage: %s [options] file1.dat [file2.dat ...]\n"
	  "       cat file1.dat [file2.dat ...] | %s [options]\n"
	  "Options (n = integer, f = float, s = string):\n"
	  "         -date YYYYMMDD output only data written on this date\n"
	  "      -averays n        average n rays\n"
	  "     -avegates n        average n gates\n"
	  "       -prefix dir      write netCDF files in specified directory\n"
	  "  -noisefactor f        set noise rejection threshold (default 4)\n"
	  "      -betacal f        multiply beta by f x 1.0e-7 to get correct value\n"
	  "      -noclean          don't clean noisy rays\n"
	  "         -user s        name and email address of the user\n"
	  "     -location s f f f  location of the lidar (e.g. Chilbolton), followed by\n"
	  "                          its latitude (N), longitude (E) and altitude (m)\n"
	  "         -site s        use known lat, lon and alt for site s - currently\n"
	  "                           understood sites are Chilbolton and Cabauw\n"
	  "      -comment s        comment to be added to each file\n"
	  "      -verbose          talk to me\n", argv[0], argv[0]);
  exit(0);
}

int
main(int argc,char **argv)
{
  LCTime time;
  LCCloudBases cloudbases;
  LCBackscatterProfile profile_hr,profile;
  char *tmp_name;
  char aname[512], *prefix=".";
  char units[] = "(sr m)^(-1)";
  char units_html[] = "sr<sup>-1</sup> m<sup>-1</sup>";
#ifdef CHIL_USE_OLD_NAMES
  char suffix[] = "lc";
#else
  char suffix[] = "ct75k";
#endif
  char join[2] = "\0\0";
  char *user = NULL;
  char *location = NULL;
  char *comment = NULL;
  char *llocation = NULL;
  float tmp_float, heights[3];
  float decimalhour, lastdecimalhour=-1, stdn, noisefactor=4.0;
  float azimuth = 0, elevation = 90, betacal = 1.0;
  float latitude, longitude, altitude;
  pid_t pid = 0;
  size_t index=0, tmp_index;
  size_t start[2], count[2];
  int averays=1, avegates=1, rays_read, cloudgate;
  int rayid, gateid, wlid, altid, latid, lonid;
  int ncid, rangeid, timeid, betaid, betarawid, stdnid;
  int betanoisyid, basesid, base1id, base2id, base3id;
  int elevationid, azimuthid;
  int dimids[2], ngates=0;
  short int beta_raw[LC_MAX_GATES];
  short int tmp_word;
  unsigned short int iarg=0,r_year=0,r_month=0,r_day=0;
  char bydate=0;
  char verbose=0, file_open=0, quiet=0, cleanray=1;
  char fixed_elevation = 0, fixed_azimuth = 0;
  char betacal_override = 0;

  /* Process command-line arguments */

  while (++iarg < argc) {
    if (*argv[iarg] != '-') {
      /* Found a ceilometer filename. The next function is cunning - it makes
       * stdin read the contents of a list of files. */
      if (!(pid = files2stdin(argc-iarg, argv+iarg))) {
	ERROR("Failed to read files", 1);
      }
      break;
    }
    else if (strcmp(argv[iarg],"-date") == 0 && iarg+1<argc) {
      bydate=1;
      if (sscanf(argv[++iarg],"%04hu%02hu%02hu",
		 &r_year,&r_month,&r_day) < 3) {
	ERROR("Failed to read date: should be in the form YYYYMMDD",1);
      }
    }
    else if (strcmp(argv[iarg],"-avegates") == 0 && iarg+1<argc) {
      avegates = atoi(argv[++iarg]);
      if (avegates < 1) {
	WARNING("Cannot average negative or zero gates");
	avegates = 1;
      }
    }
    else if (strcmp(argv[iarg],"-averays") == 0 && iarg+1<argc) {
      averays = atoi(argv[++iarg]);
      if (averays < 1) {
	WARNING("Cannot average negative or zero rays");
	averays = 1;
      }
    }
    else if (strcmp(argv[iarg],"-noisefactor") == 0 && iarg+1<argc) {
      noisefactor=atof(argv[++iarg]);
    }
    else if (strcmp(argv[iarg],"-betacal") == 0 && iarg+1<argc) {
      betacal = atof(argv[++iarg]);
      if (betacal <= 0.0) {
	ERROR("Multiplicative calibration factor cannot be negative or zero", 1);
      }
      betacal_override = 1;
    }
    else if (strcmp(argv[iarg],"-noclean") == 0) {
      cleanray=0;
    }
    else if (strcmp(argv[iarg],"-prefix") == 0 && iarg+1<argc) {
      prefix=argv[++iarg];
    }
    else if (strcmp(argv[iarg],"-user") == 0) {
      if (argc > iarg+1) {
	if (user) {
	  WARNING("User specified more than once");
	}
	user = argv[++iarg];
      }
      else {
	ERROR("User not specified",1);
      }
    }
    else if (strcmp(argv[iarg],"-comment") == 0) {
      if (argc > iarg+1) {
	if (comment) {
	  WARNING("One of the comments will not appear");
	}
	comment = argv[++iarg];
      }
      else {
	ERROR("Comment not specified", 1);
      }
    }
    else if (strcmp(argv[iarg],"-location") == 0) {
      if (argc > iarg+4) {
	if (location) {
	  WARNING("Location specified more than once");
	}
	location = argv[++iarg];
	latitude = atof(argv[++iarg]);
	longitude = atof(argv[++iarg]);
	altitude = atof(argv[++iarg]);
      }
      else {
	ERROR("Not all location arguments were provided", 1);
      }
    }
    else if (strcmp(argv[iarg], "-site") == 0) {
      if (argc > iarg+1) {
	++iarg;
	if (strcasecmp(argv[iarg], "Chilbolton") == 0) {
	  location = "Chilbolton";
	  latitude = CHIL_LAT;
	  longitude = CHIL_LON;
	  altitude = CHIL_ALT;
	}
	else if (strcasecmp(argv[iarg], "Cabauw") == 0) {
	  location = "Cabauw";
	  latitude = CABAUW_LAT;
	  longitude = CABAUW_LON;
	  altitude = CABAUW_ALT;
	}
	else {
	  WARNING("Site unknown: I only know about Chilbolton and Cabauw");
	}
      }
    }
    else if (strcmp(argv[iarg], "-elevation") == 0) {
      if (argc > iarg+1) {
	elevation = atof(argv[++iarg]);
	fixed_elevation = 1;
      }
      else {
	ERROR("Elevation not specified",1);
      }
    }
    else if (strcmp(argv[iarg], "-azimuth") == 0) {
      if (argc > iarg+1) {
	azimuth = atof(argv[++iarg]);
	fixed_azimuth = 1;
      }
      else {
	ERROR("Azimuth not specified",1);
      }
    }
    else if (strcmp(argv[iarg],"-verbose") == 0) {
      verbose=1;
      quiet=0;
    }
    else if (strcmp(argv[iarg],"-quiet") == 0) {
      quiet=1;
      verbose=0;
    }

    else {
      fprintf(stderr, "Error: `%s' not understood or not accompanied "
	      "by enough subsequent arguments\n", argv[iarg]);
      Usage(argv);
    }
  }

  if (!location) {
    WARNING("Location unknown: latitude, longitude and altitude will not be written");
    llocation = "";
  }
  else {
    char *ch = strdup(location);
    llocation = ch;
    join[0] = '-';
    while (*ch) {
      *ch = tolower(*ch);
      ++ch;
    }
  }

  while (1) {
    rays_read = (LCGetAveragedRay(stdin,&time,&cloudbases,&profile_hr,
			       averays));
    if (rays_read) {
      /* PROCESS DATA */
      LCAverageRay(&profile_hr,&profile,avegates);
      if (cleanray) {
	/* First save raw profile */
	memcpy(beta_raw, profile.beta, sizeof(*profile.beta)*profile.ngates);
	LCCustomCleanRay(&cloudgate,&stdn,&profile,8,noisefactor);
      }
      decimalhour = ((float) time.hour) + ((float) time.minute)/60.0 +
	((float) time.second)/3600.0;
    }

    if (file_open && (!rays_read || (lastdecimalhour>decimalhour))) {
      /* CLOSE FILE */
      handle_error(nc_close(ncid));
      index = file_open = 0;
      if (bydate) {
	bydate = 2;
      }
      if (verbose) {
	fprintf(stderr, "\n");
      }
    }

    if (!rays_read || bydate == 2) {
      /* EXIT */
      exit(0);
    }

    if (!file_open && !(bydate && (r_year != time.year 
				  || r_month != time.month
				  || r_day != time.day))) {
      char *version = CHIL_VERSION;

      /* OPEN NEW FILE */
      sprintf(aname, 
	      "%s/%04hu%02hu%02hu_%s%s%s.nc",
	      prefix, time.year, time.month, time.day,
	      llocation, join, suffix);
      /* Open netCDF file for writing */
      handle_error(nc_create(aname, NC_CLOBBER, &ncid));
      index = 0;
      if (!quiet)
	fprintf(stderr, "Writing %s\n", aname);

      /* Add some global attributes */
      if (location) {
	handle_error(nct_append_attribute(ncid, NC_GLOBAL, "location",
					  location, " "));
      }

      tmp_name = "CT75K";
      handle_error(nc_put_att_text(ncid, NC_GLOBAL, "system",
				   strlen(tmp_name), tmp_name));

      if (location) {
	nct_append_attribute(ncid, NC_GLOBAL, "title", location, " ");
      }
      tmp_name = "Vaisala 905-nm CT75K lidar ceilometer";
      nct_append_attribute(ncid, NC_GLOBAL, "title", tmp_name, " ");
 
      tmp_word = time.day;
      handle_error(nc_put_att_short(ncid, NC_GLOBAL, "day", NC_SHORT,
				    1, &(tmp_word)));

      tmp_word = time.month;
      handle_error(nc_put_att_short(ncid, NC_GLOBAL, "month", NC_SHORT,
				    1, &(tmp_word)));

      tmp_word = time.year;
      handle_error(nc_put_att_short(ncid, NC_GLOBAL, "year", NC_SHORT,
				    1, &(tmp_word)));

      if (comment) {
	handle_error(nct_append_attribute(ncid, NC_GLOBAL, "comments",
					  comment, " "));
      }


      /* Add a `history' attribute */
      handle_error(nct_add_history(ncid, "Generated", user));

      /* Add a `command_line' attribute */
      handle_error(nct_add_command_line(ncid, argc, argv));

      handle_error(nc_put_att_text(ncid, NC_GLOBAL, "software_version",
				   strlen(version), version));
 
      /* Add time/ray dimension */
      handle_error(nc_def_dim(ncid, "time", NC_UNLIMITED, &rayid));

      /* Add range/gate dimension  */
      ngates = profile.ngates;
      handle_error(nc_def_dim(ncid, "range", ngates, &gateid));

      /* Add some scalar variables */
      handle_error(nc_def_var(ncid, "wavelength", NC_FLOAT, 0, dimids, 
			      &wlid));
      add_attributes(ncid, wlid, "nm", NULL, "Lidar wavelength");

      if (location) {
	handle_error(nc_def_var(ncid, "latitude", NC_FLOAT, 0, dimids, 
				&latid));
	add_attributes(ncid, latid, "degrees_north", NULL, NULL);

	handle_error(nc_def_var(ncid, "longitude", NC_FLOAT, 0, dimids, 
				&lonid));
	add_attributes(ncid, lonid, "degrees_east", NULL, NULL);

	handle_error(nc_def_var(ncid, "altitude", NC_FLOAT, 0, dimids, 
				&altid));
	add_attributes(ncid, altid, "m", NULL,
		       "Altitude of lidar above mean sea level");
      }

      if (fixed_elevation) {
	handle_error(nc_def_var(ncid, "elevation", NC_FLOAT, 0, dimids,
				&elevationid));
	add_attributes(ncid, elevationid, "degrees", NULL,
		       "Elevation above horizon (fixed)");
      }
      if (fixed_azimuth) {
	handle_error(nc_def_var(ncid, "azimuth", NC_FLOAT, 0, dimids,
				&azimuthid));
	add_attributes(ncid, azimuthid, "degrees", NULL,
		       "Azimuth clockwise from due north (fixed)");
      }

      /* Add time coordinate variable */ 
      dimids[0] = rayid;
      handle_error(nc_def_var(ncid, "time", NC_FLOAT, 1, dimids,
			      &timeid));
      add_attributes(ncid, timeid, "hours UTC", NULL,
		     "Decimal hours since midnight");

      /* Add range coordinate variable */
      dimids[0] = gateid;
      handle_error(nc_def_var(ncid, "range", NC_FLOAT, 1, dimids, &rangeid));
      add_attributes(ncid, rangeid, "km", NULL, "Range from the lidar");

      /* Add cloud bases and stdn parameter */
      dimids[0] = rayid;
      handle_error(nc_def_var(ncid, "bases", NC_BYTE, 1, dimids,
			      &basesid));
      tmp_name = "Number of cloud bases";
      handle_error(nc_put_att_text(ncid, basesid, "long_name",
				   strlen(tmp_name), tmp_name));

      tmp_float = NO_SIGNAL;
      handle_error(nc_def_var(ncid, "base1", NC_FLOAT, 1, dimids,
			      &base1id));
      add_attributes(ncid, base1id, "km", NULL,
		     "Range of the first cloud base from the lidar");
      handle_error(nc_put_att_float(ncid, base1id, "missing_value",
				    NC_FLOAT, 1, &tmp_float));

      handle_error(nc_def_var(ncid, "base2", NC_FLOAT, 1, dimids,
			      &base2id));
      add_attributes(ncid, base2id, "km", NULL,
		     "Range of the second cloud base from the lidar");
      handle_error(nc_put_att_float(ncid, base2id, "missing_value",
				    NC_FLOAT, 1, &tmp_float));

      handle_error(nc_def_var(ncid, "base3", NC_FLOAT, 1, dimids,
			      &base3id));
      add_attributes(ncid, base3id, "km", NULL,
		     "Range of the third cloud base from the lidar");
      handle_error(nc_put_att_float(ncid, base3id, "missing_value",
				    NC_FLOAT, 1, &tmp_float));

      if (cleanray) {
	handle_error(nc_def_var(ncid, "stdn", NC_FLOAT, 1, dimids,
				&stdnid));
	add_attributes(ncid, stdnid, units, units_html,
		       "Standard deviation of the backscatter coefficient "
		       "background noise");
	add_scale_factor(ncid, stdnid, betacal);
      }

      /* Add backscatter coefficient */
      dimids[0] = rayid;
      dimids[1] = gateid;

      handle_error(nc_def_var(ncid, "beta", NC_SHORT, 2,
			      dimids, &betaid));
      add_attributes(ncid, betaid, units, units_html,
		     "Attenuated bacscatter coefficient");

      add_scale_factor(ncid, betaid, betacal);

      if (cleanray) {
	tmp_word = 0;
	handle_error(nc_put_att_short(ncid, betaid, "missing_value",
				      NC_SHORT, 1, &tmp_word));
	nct_append_attribute(ncid, betaid, "comments",
	     "Speckle noise has been removed, but note that when speckle "
	     "noise is strong its removal "
	     "reduces the effective sensitivity of the instrument.", " ");

	handle_error(nc_def_var(ncid, "beta_raw", NC_SHORT, 2,
				dimids, &betarawid));
	add_attributes(ncid, betarawid, units, units_html,
		       "Raw attenuated bacscatter coefficient");
	add_scale_factor(ncid, betarawid, betacal);
	betanoisyid = betarawid;
      }
      else {
	betanoisyid = betaid;
      }

      nct_append_attribute(ncid, betanoisyid, "comments",
	   "The data can be affected by speckle "
	   "noise in cloud-free regions.", " ");

      if (betacal_override) {
	nct_append_attribute(ncid, NC_GLOBAL, "comments",
	     "Calibration has been performed by utilising the fact "
	     "that the integral of the attenuated backscatter coefficient "
	     "through thick stratocumulus is equal to 1/(2k), where k is "
	     "the extinction-to-backscatter ratio and is close to 18.75 "
	     "sr for liquid cloud droplets. We have also taken multiple "
	     "scattering into account - for the Vaisala instrument it "
	     "has an effect of around 20%. The scale_factor "
	     "attribute indicates the resulting calibration value. The "
	     "raw system calibration can be obtained by using a scale_factor "
	     "of 1.0e-7.", " ");
      }

      if (comment) {
	nct_append_attribute(ncid, NC_GLOBAL, "comments", comment, " ");
      }

      /* End define mode */
      handle_error(nc_enddef(ncid));

      /* Set the scalar variables */
      tmp_index = 0;
      tmp_float = LC_WAVELENGTH;
      handle_error(nc_put_var1_float(ncid, wlid, &tmp_index, &tmp_float));

      if (location) {
	handle_error(nc_put_var1_float(ncid, altid, &tmp_index, &altitude));
	handle_error(nc_put_var1_float(ncid, latid, &tmp_index, &latitude));
	handle_error(nc_put_var1_float(ncid, lonid, &tmp_index, &longitude));
      }

      if (fixed_elevation) {
	handle_error(nc_put_var1_float(ncid, elevationid,
				       &tmp_index, &elevation));
      }
      if (fixed_azimuth) {
	handle_error(nc_put_var1_float(ncid, azimuthid,
				       &tmp_index, &azimuth));
      }

      /* Set the range */
      for (tmp_index = 0; tmp_index < ngates; ++tmp_index) {
	tmp_float = profile.range[tmp_index] * 0.001;
	handle_error(nc_put_var1_float(ncid, rangeid,
				       &tmp_index, &tmp_float));
      }
      file_open = 1;
    }
   
    if (file_open) {
      /* WRITE RAY */
      handle_error(nc_put_var1_float(ncid, timeid, &index,
				     &decimalhour));
      handle_error(nc_put_var1_int(ncid, basesid, &index,
				   &(cloudbases.number)));

      heights[2] = heights[1] = heights[0] = NO_SIGNAL;
      if (cloudbases.number > 0) {
	heights[0] = cloudbases.height1 * 0.001;
	if (cloudbases.number > 1) {
	  heights[1] = cloudbases.height2 * 0.001;
	  if (cloudbases.number > 2) {
	    heights[2] = cloudbases.height3 * 0.001;
	  }
	}
      }

      handle_error(nc_put_var1_float(ncid, base1id, &index,
				   heights));
      handle_error(nc_put_var1_float(ncid, base2id, &index,
				   heights+1));
      handle_error(nc_put_var1_float(ncid, base3id, &index,
				   heights+2));

      if (cleanray) {
	handle_error(nc_put_var1_float(ncid, stdnid, &index,
				       &stdn));
      }
      start[0] = index;
      start[1] = 0;
      count[0] = 1;
      count[1] = ngates;
      handle_error(nc_put_vara_short(ncid, betaid, start,
				     count, profile.beta));
      if (cleanray) {
	handle_error(nc_put_vara_short(ncid, betarawid, start,
				       count, beta_raw));
      }

      ++index;
      if (verbose) {
	fprintf(stderr, ".");
      }
    }
    lastdecimalhour=decimalhour;
  }
  exit(0);
}
