#include <stdio.h>
#include <stdlib.h>
#include <netcdf.h>
#include "cfft.h"
#include "chil.h" /* Just for the lon, lat & alt at the moment */
#include "version.h"

#define FILE_PREFIX ""
#define FILE_SUFFIX "_94fft.nc"

int ncstatus;
#define handle_error(a) if ((ncstatus = (a)) != NC_NOERR) { \
    fprintf(stderr, "NetCDF error on line %d: %s\n", __LINE__, \
            nc_strerror(ncstatus)); \
    exit(-1); \
}

void
add_attributes(int ncid, int fieldid, char *units, char *long_name)
{
  if (units) {
    handle_error(nc_put_att_text(ncid, fieldid, "units",
				 strlen(units), units));
  }
  if (long_name) {
    handle_error(nc_put_att_text(ncid, fieldid, "long_name",
				 strlen(long_name), long_name));
  }
}

void
Usage(char **argv)
{
  fprintf(stderr, "Usage: %s [options] file1.fft [file2.fft ...]\n"
	  "Options:\n"
	  "     -reversev          negate the velocities\n"
	  "    -noprocess          don't subtract noise\n"
	  "      -noclean          don't remove spike\n"
	  "       -linear          use arbitrary linear units\n"
	  "  -rangeoffset f        range offset in metres\n",
	  argv[0]);
  exit(0);
}

int
main(int argc, char **argv)
{
  unsigned long flags = CFFT_USE_LOG_SNR
    | CFFT_REMOVE_NOISE | CFFT_REMOVE_SPIKE;
  size_t profile = 0;
  CFFTData data;
  char *filename;
  char *prefix = ".";
  char ncfilename[512];
  int iarg = 1;
  float range_offset = 0;
  short last_day, last_month, last_year;

  /* NetCDF IDs */
  int ncid, rangeid, vid, profileid, gateid, binid;
  int latid, lonid, altid, specid, freqid, timeid, noiseid;
  int dimids[3];

  /* flags */
  char reversev;
  char newfile = 1;
  char quiet = 0;

  if (argc < 2) {
    Usage(argv);
  }

  while (iarg < argc && *argv[iarg] == '-') {
    if (strcmp(argv[iarg], "-reversev") == 0) {
      reversev = 1;
      flags |= CFFT_REVERSE_V;
    }
    else if (strcmp(argv[iarg], "-noclean") == 0) {
      flags &= ~CFFT_REMOVE_SPIKE;
    }
    else if (strcmp(argv[iarg], "-quiet") == 0) {
      quiet = 1;
    }
    else if (strcmp(argv[iarg], "-noprocess") == 0) {
      flags &= ~CFFT_REMOVE_NOISE;
    }
    else if (strcmp(argv[iarg], "-linear") == 0) {
      flags &= ~CFFT_USE_LOG_SNR;
    }
    else if (strcmp(argv[iarg], "-rangeoffset") == 0) {
      if (argc <= iarg+1) {
	fprintf(stderr, "Failed to read range offset\n");
	exit(-1);
      }
      range_offset = atof(argv[++iarg]);
    }
    else {
      fprintf(stderr, "Error: `%s' not understood\n", argv[iarg]);
      Usage(argv);
    }
    ++iarg;
  }

  if (!quiet) {
    cfft_verbose();
  }

  cfft_set_processing_flags(&data, flags);
  cfft_set_range_offset(&data, range_offset);

  while (iarg < argc) {
    int index = 0; /* Index of the current profile in the input file */

    filename = argv[iarg];
    if (cfft_open(filename, &data) != CFFT_OK) {
      fprintf(stderr, "Error reading %s\n", filename);
      exit(2);
    }

    data.spikebleed = 1;
    cfft_set_noise_processing(&data, 50, 3);

    while (cfft_read_profile(&data, index) == CFFT_OK) {
      size_t tmp_index;
      size_t start[3], count[3];
      char *tmp_name;
      float tmp_float;
      char *version = CHIL_VERSION;
      cfft_process_spectra(&data);

      if (!newfile && (last_day != data.day || last_month != data.month
		       || last_year != last_year)) {
	handle_error(nc_close(ncid));
	fprintf(stderr, "\n");
	newfile = 1;
      }

      if (newfile) {
	/* Open netCDF file for writing */
	sprintf(ncfilename, 
		"%s/" FILE_PREFIX "%04hu%02hu%02hu" FILE_SUFFIX,
		prefix, CHILTrueYear(data.year), data.month, data.day);
	handle_error(nc_create(ncfilename, NC_CLOBBER, &ncid));
	fprintf(stderr, "Writing %s: profile", ncfilename);

	/* Add some global attributes */
	tmp_name = "Galileo";
	handle_error(nc_put_att_text(ncid, NC_GLOBAL, "system",
				     strlen(tmp_name), tmp_name));

	last_day = data.day;
	handle_error(nc_put_att_short(ncid, NC_GLOBAL, "day", NC_SHORT,
				      1, &last_day));

	last_month = data.month;
	handle_error(nc_put_att_short(ncid, NC_GLOBAL, "month", NC_SHORT,
				      1, &last_month));

	last_year = CHILTrueYear(data.year);
	handle_error(nc_put_att_short(ncid, NC_GLOBAL, "year", NC_SHORT,
				      1, &last_year));

	handle_error(nc_put_att_text(ncid, NC_GLOBAL, "software_version",
				     strlen(version), version));

	/* Add time/profile dimension */
	handle_error(nc_def_dim(ncid, "time", NC_UNLIMITED, &profileid));

	/* Add range/gate dimension  */
	handle_error(nc_def_dim(ncid, "range", data.ngates, &gateid));

	/* Add velocity/bin dimension  */
	handle_error(nc_def_dim(ncid, "velocity", data.nbins, &binid));

	/* Add some scalar variables */
	handle_error(nc_def_var(ncid, "frequency", NC_FLOAT, 0, dimids, 
				&freqid));
	add_attributes(ncid, freqid, "GHz", "Radar frequency");

	handle_error(nc_def_var(ncid, "latitude", NC_FLOAT, 0, dimids, 
				&latid));
	add_attributes(ncid, latid, "degrees_north", NULL);

	handle_error(nc_def_var(ncid, "longitude", NC_FLOAT, 0, dimids, 
				&lonid));
	add_attributes(ncid, lonid, "degrees_east", NULL);

	handle_error(nc_def_var(ncid, "altitude", NC_FLOAT, 0, dimids, 
				&altid));

	add_attributes(ncid, altid, "m",
		       "Altitude of antenna above mean sea level");

	/* Add time coordinate variable */ 
	dimids[0] = profileid;
	handle_error(nc_def_var(ncid, "time", NC_FLOAT, 1, dimids,
				&timeid));
	add_attributes(ncid, timeid, "hours UTC", "Decimal hours since midnight");

	handle_error(nc_def_var(ncid, "noise", NC_FLOAT, 1, dimids,
				&noiseid));
	add_attributes(ncid, noiseid, "arbitrary", "Linear mean noise level");

	/* Add range coordinate variable */
	dimids[0] = gateid;
	handle_error(nc_def_var(ncid, "range", NC_FLOAT, 1, dimids, &rangeid));
	add_attributes(ncid, rangeid, "km", "Range from antenna");

	/* Add velocity coordinate variable */
	dimids[0] = binid;
	handle_error(nc_def_var(ncid, "velocity", NC_FLOAT, 1, dimids, &vid));
	if (flags & CFFT_REVERSE_V) { 
	  add_attributes(ncid, vid, "m/s", "Velocity towards antenna");
	}
	else {
	  add_attributes(ncid, vid, "m/s", "Velocity away from antenna");
	}
      
	dimids[0] = profileid;
	dimids[1] = gateid;
	dimids[2] = binid;

	tmp_float = CFFT_NO_SIGNAL;
	handle_error(nc_def_var(ncid, "spec", NC_FLOAT, 3, dimids, &specid));
	if (flags & CFFT_USE_LOG_SNR)
	  add_attributes(ncid, specid, "dB", "Spectral density signal-to-noise ratio");
	else {
	  add_attributes(ncid, specid, "arbitrary", "Linear spectral density");
	}
	handle_error(nc_put_att_float(ncid, specid, "missing_value",
				      NC_FLOAT, 1, &tmp_float));

	/* End define mode */
	handle_error(nc_enddef(ncid));

	/* Set the scalar variables */
	tmp_index = 0;
	tmp_float = GALILEO_FREQ;
	handle_error(nc_put_var1_float(ncid, freqid, &tmp_index, &tmp_float));

	tmp_float = CHIL_ALT;
	handle_error(nc_put_var1_float(ncid, altid, &tmp_index, &tmp_float));

	tmp_float = CHIL_LAT;
	handle_error(nc_put_var1_float(ncid, latid, &tmp_index, &tmp_float));

	tmp_float = CHIL_LON;	       
	handle_error(nc_put_var1_float(ncid, lonid, &tmp_index, &tmp_float));

	/* Set the range */
	handle_error(nc_put_var_float(ncid, rangeid, data.range));

	/* Set the velocity */
	handle_error(nc_put_var_float(ncid, vid, data.velocity));

	newfile = 0;
	profile = 0; /* Index of the current profile in the output file */
      }

      /* Add a new profile of spectra */
      start[0] = profile;
      start[1] = 0;
      start[2] = 0;

      count[0] = 1;
      count[1] = data.ngates;
      count[2] = data.nbins;

      fprintf(stderr, " %d", profile);

      tmp_index = profile;
      tmp_float = ((float) data.hour) + data.minute / 60.0 + data.second / 3600.0;
      handle_error(nc_put_var1_float(ncid, timeid, &tmp_index, &tmp_float));
      handle_error(nc_put_var1_float(ncid, noiseid, &tmp_index, &(data.mean_noise)));
      handle_error(nc_put_vara_float(ncid, specid, start, count, data.spectra));
      ++profile;
      ++index;
    }
    cfft_close(&data);

    ++iarg;
  }
  if (!newfile) {
    handle_error(nc_close(ncid));
    fprintf(stderr, "\n");
  }
  exit(0);
}
